#!/usr/bin/perl -w

package Data::Array2ArrayMap;
use Carp;
use strict;

use vars qw($VERSION);
$VERSION = '0.12';

=head1 NAME

Data::Array2ArrayMap - perl interface for hierarchical data storage

=head1 SYNOPSIS

This help is only on perl interface of Data::Array2ArrayMap.

First you 

For example, 

=head1 DESCRIPTION

L<Hash::MultiKey>,
L<Tie::ListKeyedHash>,
L<Data::Hive>,
L<DBM::Deep>,

=head1 METHODS

=cut

=head3 new()

new(backend-specific arguments) is generally backend-specific.

=over 4

=item Common Options

=over 4 

=item *

strict - if set to 0 the module will allow 

=back

=back

=cut

sub new {
    my $class  = shift;
    my $self = {};
    bless($self,$class);
    return $self;
}


=head2 Query Capabilities Methods

Those methods allow quering of backend's capabilities.

=head3 is_readonly()

=cut
sub is_readonly() {return 1}

=head3 is_persistent_storage()

=cut
sub is_persistent_storage() {return 0}

=head3 can_store_empty_array()

=cut
sub can_store_empty_array() {return 0}

=head3 has_find()

=cut
sub has_find() {return 0}

=head3 has_clone()

=cut
sub has_clone() {return 0}

=head3 can_list_paths()

=cut
sub can_list_paths() {return 0}

=head3 has_keys_at()

=cut
sub has_keys_at() {return 0}


=head2 Data Access Methods

=head3 get()

get([path]) - fetch array of data stored at given path.
In array context returns the stored array.
In scalar context returns first element of the array.

Examples: 
my $first_val=$map->get([100,"July"]);
my @vals=$map->get([100,"July"]);
my @vals2=$map->get(\@path2);

=cut
sub get {
    my ($self, $pathptr)=@_;
    if (wantarray) {
	return @{$self->get_ref($pathptr)};
    } else {
	return $self->get_ref($pathptr)->[0];
    }
}

=head3 get_ref()

get_ref([path]) - the same as get([path]), but returns the reference to the array.
for efficiency reasons for some backends the referenced array may be not a copy of stored data, but the
data itself, so the got reference should be treated read-only.

=cut

sub get_ref_impl_with_get {
    my ($self, $pathptr)=@_;
    return [$self->get($pathptr)];
}

sub get_ref_impl_with_if {
    my ($self, $pathptr)=@_;
    my $ret = $self->get_ref_if($pathptr);
    return $ret? $ret : [];
}

*get_ref=*get_ref_impl_with_if;

=head3 get_ref_if()

get_ref_if([path]) - returns reference to array of data at given path
or undef if there was nothing stored. It tries to show the difference between
stored empty array and nothing stored. This difference is backend-specific:
Some backends feel the difference between stored empty array and stored nothing,
some does not.

=cut
sub get_ref_if {
    my ($self, $pathptr)=@_;
    my $ref=$self->get_ref($pathptr);
    if ($#{$ref}==0) {
	return undef;
    } else {
	return $ref;
    }
}

=head3 set()

set([path],[values]) - store array of data at given path.
It is designed to store scalar values.
note that for some backends (it depends on the backend implementation)
the perl-specific data such as object or subroutine references
can be stored incorrectly.

return value is backend-specific.

Examples: 
$map->set([100,"July"], [1]);
$map->set([100,"July"], \@data);
$map->set(\@path2, \@data2);

=cut
sub set {
    my ($self, $pathptr, $valueptr )= @_;
    Carp::croak "Array2ArrayMap: interface stub called for set";
}

=begin comment
=head3 put()

put([path],value) - store single value at given path.
return value is backend-specific.

It is method implemented directly in interface for convenience and compatibility reasons
(namely, for compatibility with Tie::ListKeyedHash).
Not recommended for use, as it can be a bit slower than native set() methods.

Examples: 
$map->put([100,"July"], 1);
$map->put(\@path, "Fog");

=end comment

=cut
sub put {
    my ($self, $pathptr, $value)= @_;
    return $self->set($pathptr,[$value]);
}

sub append {
    my ($self, $pathptr, $valueptr)= @_;
    my $data_ref=$self->get_ref($pathptr);
    push @$data_ref, @$valueptr;
    return $self->set($pathptr, $data_ref);
}

sub _subtract_arrays {
    my %hash=map {$_=>1} @{$_[1]};
    return grep {!$hash{$_}} @{$_[0]};
}

# sub _intersect_arrays {
#     my %hash=map {$_=>1} @{shift()};
#     return [grep {$hash{$_}} @{shift()}];
# }
# sub _unite_arrays {
#     my %tmp_hash=map {$_=>1} @{shift()}, @{shift()};
#     return [keys (%tmp_hash)];
# }

sub unique_append {
    # is a shortcut for remove+append
    my ($self, $pathptr, $valueptr)= @_;
    my $data_ref=$self->get_ref($pathptr);
    #    this implementation preserves the order of values
    my @newdata=_subtract_arrays($data_ref, $valueptr);
    push @newdata, @$valueptr;
    return $self->set($pathptr, \@newdata);
}

sub remove {
    my ($self, $pathptr,$valueptr)=@_;
    my $data_ref=$self->get_ref($pathptr);
    my @newdata=_subtract_arrays($data_ref, $valueptr);
    return $self->set($pathptr, \@newdata);
}

sub find {
    my ($self, $pathptr)=@_;
    Carp::croak "Array2ArrayMap: interface stub called for find";
}

#sub delete {
#    my ($self, $pathptr)=@_;
#    Carp::croak "Array2ArrayMap: interface stub called";
#}

# keys_at returns all keys at given path' level
sub keys_at {
    my ($self, $pathptr) = @_;
    Carp::croak "Array2ArrayMap: interface stub called for keys_at";
}

# returns paths that starts with [path]
sub paths_prefix {
    my ($self, $pathptr) = @_;
    my @pathlist = $self->subpaths_at($pathptr);
    map {unshift @$_, @$pathptr} @pathlist;
    return @pathlist;
}

# returns paths that ends with [path]
sub paths_suffix {
    Carp::croak "Array2ArrayMap: interface stub called for path_suffix";
}

# returns all paths
sub paths_all {
    Carp::croak "Array2ArrayMap: interface stub called for path_suffix";
}

# list subpaths that starts with $pathptr
sub subpaths_at {
    my ($self, $pathptr) = @_;
    my $shiftlen=@$pathptr;
    my @paths_at=$self->paths_prefix($pathptr);
    foreach (@paths_at) {
	for (my $i=$shiftlen;$i>0;$i--) {shift @$_}
    }
    return @paths_at;
}

# compatibility with Tie::ListKeyedHash; and tie HASH
# Not recommended for use
sub clearTOREMOVESUFFIXAFTERTEST {
    return &clear_prefix([]);
}

sub delete {
    return &clear_prefix;
}

sub clear_prefix {
    my ($self, $pathptr)=@_;
    foreach ($self->paths_prefix($pathptr)) {
	$self->delete($_);
    }
}

sub clear_suffix {
    my ($self, $pathptr)=@_;
    foreach ($self->paths_suffix($pathptr)) {
	$self->delete($_);
    }
}

# TODO: rename and implement!!!

## arguments path,object: deletes object on the branch after path
## hack - indicates code that needs special cache without find.
## ( notobject->attr->context1...->contextn)
## but rather attr->object[->join_context] (verify that 2 args!) by interface
sub all_clear {
    &delete;
}

#sub clone {
#    my $self=shift;
#    Carp::croak "Array2ArrayMap: interface stub called";
#}

########################## TIE::HASH ###############################

sub TIEHASH {
    bless {}, shift;
}

sub CLEAR {
    shift()->clear();
}

sub FETCH {
    my ($self, $keys) = @_;
    $keys = [$keys eq '' ? ('') : split /$;/, $keys, -1] unless ref $keys eq 'ARRAY';#/
    @$keys or croak "Empty multi-key\n";
    $self->get_ref_if($keys);
}

sub STORE {
    my ($self, $keys, $value) = @_;
    $keys = [$keys eq '' ? ('') : split /$;/, $keys, -1] unless ref $keys eq 'ARRAY';#/
    @$keys or croak "Empty multi-key\n";
    $self->rawput($keys,$value);
}

sub DELETE {
    my ($self, $keys) = @_;
    $keys = [$keys eq '' ? ('') : split /$;/, $keys, -1] unless ref $keys eq 'ARRAY';#/
    @$keys or croak "Empty multi-key\n";
    $self->delete($keys);
}

sub EXISTS {
    my ($self, $keys) = @_;
    $keys = [$keys eq '' ? ('') : split /$;/, $keys, -1] unless ref $keys eq 'ARRAY';#/
    @$keys or croak "Empty multi-key\n";
    scalar $self->get_ref_if($keys);
}

sub FIRSTKEY {
    my ($self) = @_;
    Carp::croak "Array2ArrayMap: interface stub called";
}

sub NEXTKEY {
    my ($self) = @_;
    Carp::croak "Array2ArrayMap: interface stub called";
}

# stub for tie() 
sub SCALAR {1}

1;


__END__

=head1 BUGS

None yet :)

When submitting bug reports, be sure to include full details,
including the VERSION of the module, a test script and a test data
demonstrating the problem!

=head1 CREDITS

to Alexander Strelets, sav@imath.kiev.ua for his work on SDB backend.

=head1 WEBSITE

You can find information about Data::Array2ArrayMap at:

   http://html-tmpl-pro.sourceforge.net

=head1 PUBLIC GIT SERVER

Data::Array2ArrayMap now has a publicly accessible GIT repository at
http://git.altlinux.org/people/viy/ at
(http://git.altlinux.org/people/viy/packages/?p=perl-Data-Array2ArrayMap.git;a=summary).
Give it a try!

=head1 AUTHOR

I. Vlasenko, E<lt>viy@altlinux.orgE<gt>

=head1 LICENSE

  Data::Array2ArrayMap - perl interface for hierarchical data storage.
  Copyright (C) 2002-2021 I. Vlasenko, (viy@altlinux.org)

  This module is free software; you can redistribute it and/or modify it
  under the terms of either:

  a) the GNU General Public License as published by the Free Software
  Foundation; either version 2, or (at your option) any later version,

  or

  b) the "Artistic License" which comes with this module.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See either
  the GNU General Public License or the Artistic License for more details.

  You should have received a copy of the Artistic License with this
  module, in the file ARTISTIC.  If not, I'll be glad to provide one.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA

=cut
